# =========================================================================
# $Id: tm.tcl,v 1.1.2.5 2007/07/06 09:47:06 belinfan Exp $
# =========================================================================

# tree-matching library.
#
# input:
#  - list of three-tuples < <position, value> list, target >
#    to be used to build the mapping table
#
# output:
#  - a routine that maps items to targets, given
#     - a routine to compare values
#     - a routine to get a value at a position from the item
#     - a routine to get an arity at a position from the item

# original interface:
#  tm_init: (:position -> int), (:value -> int), (:target -> int) -> void
#  tm_add_pattern: <type, position, val>-list, target -> { success, failure }
#     (failure may result if the pattern introduces non-confluency)
#     (type = 'n' | 'a'   )
#     (val = name | arity )
#  tm_match: item,
#            (:item, position -> value),
#            (:item, position -> arity) -> target

# this original interface has one disadvantage: we have only one
# set of rules at a time.
# we would like to change this such that we can have multiple
# rulesets, so:
# tm_init now can be called multiple times, once for each
#      distinct set of rules, and returns a tm identifier ('handle')
# the other routines now need the handle as argument
# do we need a general cleanup routine?

proc tm_init {poscmp valuecmp targetcmp} {
	tm_setup
	set h [tm_newhandle]

#	puts [info vars]
	
	upvar #0 tm_$h tm
#	puts [info vars]
	
	set tm(nrofrules) 0
	set tm(cmp,pos) $poscmp
	set tm(cmp,value) $valuecmp
	set tm(cmp,target) $targetcmp
	set tm(match) tm_interpreted
	return $h
}

proc tm_setup {} {
	global tmadm
	if {![info exists tmadm(nrofh)]} {
		set tmadm(nrofh) 0
	}
}
proc tm_newhandle {} {
	global tmadm
	if {![info exists tmadm(nrofh)]} {
		puts stderr "tm: inialize first!"
		return -1
	}
	set h tmh$tmadm(nrofh)
	incr tmadm(nrofh)
	return $h
}

proc tm_finalize {h} {
	upvar #0 tm_$h tm
	catch { unset tm }
	return ""
}

proc tm_show {h} {
	upvar #0 tm_$h tm
		if {![info exists tm(nrofrules)]} {
			puts stderr "tm: inialize first!"
			return -1
		}
		set rulenr 0
		while {$rulenr < $tm(nrofrules)} {
			puts stderr "Rule: $tm(rule,repr,$rulenr) -> $tm(rule,target,$rulenr)"
			foreach p $tm(rule,pairs,$rulenr) {
				puts stderr "\t$p"
			}
			incr rulenr
		}
}

proc tm_list {h} {
	upvar #0 tm_$h tm
		set rules {}
		if {![info exists tm(nrofrules)]} {
			puts stderr "tm: inialize first!"
			return $rules
		}
		set rulenr 0
		while {$rulenr < $tm(nrofrules)} {
			lappend rules [list $tm(rule,pairs,$rulenr) $tm(rule,target,$rulenr)]
			incr rulenr
		}
		return $rules
}

proc tm_add_pattern { h repr pos_value_pairs target } {
	upvar #0 tm_$h tm
		if {![info exists tm(nrofrules)]} {
			puts stderr "tm: inialize first!"
			return -1
		}
		set sorted_pairs [tm_sort_pairs $h $pos_value_pairs]
		return [tm_add_sorted_pattern $h $repr $sorted_pairs $target]
}
proc tm_add_sorted_pattern { h repr pos_value_pairs target } {
	upvar #0 tm_$h tm
		if {[tm_remains_confluent $h $repr $pos_value_pairs $target]} {
			set rulenr $tm(nrofrules)
			set tm(rule,repr,$rulenr) $repr
			set tm(rule,pairs,$rulenr) $pos_value_pairs
			set tm(rule,target,$rulenr) $target
			# debugMsg "tm_add_sorted_pattern ($pos_value_pairs)($target)"
			incr tm(nrofrules)
			set tm(match) tm_interpreted
			return 1
		} else {
			return 0
		}
}

proc tm_sort_pairs {h pairs} {
	return [lsort -command [list tm_paircmp $h] $pairs]
}

proc tm_paircmp {h a b} {
	upvar #0 tm_$h tm
		set tmp [$tm(cmp,pos) [lindex $a 1] [lindex $b 1]]
		if {$tmp != 0} {
			return $tmp
		} else {
			return [tm_typecmp $h [lindex $a 0] [lindex $b 0]]
		}
}
proc tm_typecmp {h a b} {
	upvar #0 tm_$h tm
		if {($a == "a") && ($b == "n")} {
			return -1
		} elseif {($a == "n") && ($b == "a")} {
			return 1
		} else {
			return 0
		}
}

proc tm_remains_confluent { h repr pos_value_pairs target } {
	upvar #0 tm_$h tm
		set rulenr 0
		while {$rulenr < $tm(nrofrules)} {
			if {[tm_target_cmp $h $tm(rule,target,$rulenr) $target] != 0} {
				if {![tm_pattern_distinct $h $tm(rule,pairs,$rulenr) $pos_value_pairs]} {
					puts stderr "Clash: $repr -> $target"
					puts stderr "with:  $tm(rule,repr,$rulenr) -> $tm(rule,target,$rulenr)"
					return 0
				}
			}
			incr rulenr
		}
		return 1
}

proc tm_target_cmp { h t1 t2 } {
	upvar #0 tm_$h tm
		return [$tm(cmp,target) $t1 $t2]
}

proc tm_pattern_distinct { h pat1 pat2 } {
	upvar #0 tm_$h tm
		set l1 [llength $pat1]
		set l2 [llength $pat2]
		set i1 0
		set i2 0
		while { ($i1 < $l1) && ($i2 < $l2) } {
			set pos1 [lindex [lindex $pat1 $i1] 1]
			set pos2 [lindex [lindex $pat2 $i2] 1]
			set cmp [$tm(cmp,pos) $pos1 $pos2]
			if {$cmp < 0} {
				incr i1
			} elseif  {$cmp > 0} {
				incr i2
			} else {
				set typ1 [lindex [lindex $pat1 $i1] 0]
				set typ2 [lindex [lindex $pat2 $i2] 0]
				set val1 [lindex [lindex $pat1 $i1] 2]
				set val2 [lindex [lindex $pat2 $i2] 2]
				if {[tm_typecmp $h $typ1 $typ2] != 0} {
					incr i1
					incr i2
				} elseif {[$tm(cmp,value) $val1 $val2] != 0} {
					# debugMsg "tm_pattern_distinct ($pat1)($pat2): 1"
					return 1
				} else {
					incr i1
					incr i2
				}
			}
		}
		# debugMsg "tm_pattern_distinct ($pat1)($pat2): 0"
		return 0
}

proc tm_match { h item getpos getarity } {
	upvar #0 tm_$h tm
		return [$tm(match) $h $item $getpos $getarity]
}

proc tm_interpreted { h item getpos getarity } {
	upvar #0 tm_$h tm
		if {![info exists tm(nrofrules)]} {
			puts stderr "tm: inialize first!"
			return -1
		}
		tm_search_init $h
		set cmp $tm(cmp,value)
		set rulenr 0
		while {$rulenr < $tm(nrofrules)} {
			if {[tm_try_rule $h $item $rulenr $getpos $getarity]} {
				tm_search_cleanup $h
				return [list 1 $tm(rule,target,$rulenr)]
			}
			incr rulenr
		}
		tm_search_cleanup $h
		return [list 0]
}

proc tm_search_init {h} {
}
proc tm_search_cleanup {h} {
	upvar #0 tmtmp_$h tmtmp
		catch { unset tmtmp }
}

proc tm_try_rule { h item rulenr getpos getarity} {
	upvar #0 tm_$h tm
		foreach p $tm(rule,pairs,$rulenr) {
			if {![tm_try_rule_pair $h $item $p $getpos $getarity]} {
				return 0
			}
		}
		return 1
}

proc tm_try_rule_pair { h item pair getpos getarity} {
	upvar #0 tm_$h tm
		set type [lindex $pair 0]
		set pos [lindex $pair 1]
		if {[tm_typecmp $h $type a] == 0} {
			set val [lindex $pair 2]
			set item_val [tm_get $h arity $item $pos $getarity]
			if {$val == $item_val} {
				set result 1
			} else {
				set result 0
			}
		} else {
			set val [lindex $pair 2]
			set item_val [tm_get $h value $item $pos $getpos]
			if {[$tm(cmp,value) $val $item_val] == 0} {
				set result 1
			} else {
				set result 0
			}
		}
		# debugMsg "tm_try_rule_pair $item $pair: $item_val -> $result"
		return $result
}

proc tm_get { h kind item pos getkind } {
	upvar #0 tmtmp_$h tmtmp
		if {![info exists tmtmp($item,$kind,$pos)]} {
			set tmtmp($item,$kind,$pos) [$getkind $item $pos]
		}
		return $tmtmp($item,$kind,$pos)
}

# =========================================================================

# Code to compule rule-pairs
#
proc tm_compile {h} {
	upvar #0 tm_$h tm
		if {[catch {eval [tm_code $h]} foo]} {
			warningMsg $foo
			return 0
		} else {
			if {[string compare [info procs tm_compiled_$h] ""] != 0} {
				set tm(match) tm_compiled_$h
			}
			return 1
		}
}
proc tm_code {h} {
	set rules [tm_compile_rules $h [tm_list $h]]
	set body [tm_generate_code $h $rules " " " "]
	set s ""
	append s "proc tm_compiled_$h { h item getpos getarity } {\n"
	append s $body
	append s "}\n"
	return $s
}
proc tm_compile_rules {h rules} {
	if {[llength $rules] <= 0} {
		return $rules
	}
	set sorted [tm_sort_rules $h $rules]
	set firstrule [lindex $sorted 0]
	set firstpairs [lindex $firstrule 0]
	set firsttarget [lindex $firstrule 1]
	if {[llength $firstpairs] <= 0} {
		return [list t $firsttarget]
	}
	set firstpair [lindex $firstpairs 0]
	set firstpairtype [lindex $firstpair 0]
	set firstpairpos [lindex $firstpair 1]
	
	foreach r $rules {
		set pairs [lindex $r 0]
		set target [lindex $r 1]
		if {[llength $pairs] <= 0} {
			lappend tmcomptmpdefault $r
		} else {
			set pair [lindex $pairs 0]
			if {[tm_paircmp $h $firstpair $pair] == 0} {
				set val [lindex $pair 2]
				set tailpairs [lrange $pairs 1 end]
				lappend tmcomptmp($val) [list $tailpairs $target]
			} else {
				lappend tmcomptmpdefault $r
			}
		}
	}
	set res {}
	if {[info exists tmcomptmpdefault]} {
		set defpat [tm_compile_rules $h $tmcomptmpdefault]
		set resdef $tmcomptmpdefault
	} else {
		set defpat [list d ]
		set resdef {}
	}
	foreach k [array names tmcomptmp] {
		set subpat [tm_compile_rules $h [concat $tmcomptmp($k) $resdef]]
		lappend res [list  $k $subpat]
	}
	return [list $firstpairtype $firstpairpos $res $defpat ]
}

proc tm_generate_code {h rule {indent "  "} {inc "  "} } {
	# debugMsg "tm_generate_code: ($rule)"
	if {[llength $rule] <= 0 } {
		return
	}
	set type [lindex $rule 0]
	switch -- $type {
		  t {
			set target [lindex $rule 1]
			return "[set indent]return \[list  1 [list $target ]\]\n"
		} d {
			return "[set indent]return \[list  0 \]\n"
		} a {
			set routine getarity
		} n {
			set routine getpos
		} default {
			debugMsg "tm_generate_code: Unknown type: ($type)"
		}
	}
	set pos [lindex $rule 1]
	set subrules [lindex $rule 2]
	set default [lindex $rule 3]
	set subindent "[set indent][set inc]"
	set subsubindent "[set subindent][set inc]"
	set s ""
	append s "[set indent]switch -- \[\$[set routine] \$item [list $pos]\] {\n"
	foreach r $subrules {
		set val [lindex $r 0]
		set sub [lindex $r 1]
		append s "[set subindent]$val {\n"
		append s [tm_generate_code $h $sub $subsubindent $inc]
		append s "[set subindent]}\n"
	}
	append s "[set subindent]default {\n"
	append s [tm_generate_code $h $default $subsubindent $inc]
	append s "[set subindent]}\n"
	append s "[set indent]}\n"
	return $s
}

# =========================================================================

proc tm_sort_rules {h rules} {
	return [lsort -command [list tm_rulecmp $h] $rules]
}

# compare on pairs, not on result!
proc tm_rulecmp {h a b} {
	set pairs1 [lindex $a 0]
	set pairs2 [lindex $b 0]
	set l1 [llength $pairs1]
	set l2 [llength $pairs2]
	set i1 0
	set i2 0
	while { ($i1 < $l1) && ($i2 < $l2) } {
		set cmp [tm_paircmp $h [lindex $pairs1 $i1] [lindex $pairs2 $i2]]
		if {$cmp < 0} {
			return -1
		} elseif  {$cmp > 0} {
			return 1
		} else {
			incr i1
			incr i2
		}
	}
	if {$i1 < $l1} {
		return -1
	} elseif  {$i2 < $l2} {
		return 1
	} else {
		return 0
	}
}


# =========================================================================

# Testing routine.
# Note: this routine depends on the em.tcl file for the interfaces!

proc tm_test {} {
	set h1 [em_init]
	em_add_pattern $h1 "aap!noot!*" wim
	em_add_pattern $h1 "*!zus!*" schapen
	
	tm_do_test $h1 interpreted "bla!zus!mies"  [list 1 schapen]
	tm_do_test $h1 interpreted "bla!noot!mies" [list 0]
	tm_do_test $h1 interpreted "aap!noot!mies" [list 1 wim]
	tm_do_test $h1 interpreted "bla!bla!bla"   [list 0]
	
	tm_compile $h1
	
	tm_do_test $h1 compiled  "bla!zus!mies"  [list 1 schapen]
	tm_do_test $h1 compiled  "bla!noot!mies" [list 0]
	tm_do_test $h1 compiled  "aap!noot!mies" [list 1 wim]
	tm_do_test $h1 compiled  "bla!bla!bla"   [list 0]

	set h2 [em_init]
	em_add_pattern $h2 "aap!noot" jet
	em_add_pattern $h2 "*!noot" jet
	em_add_pattern $h1 "aap!noot" teun
	em_add_pattern $h1 "*!noot" teun
	
	tm_do_test $h2 interpreted "aap!noot"    [list 1 jet]
	tm_do_test $h2 interpreted "bla!noot"    [list 1 jet]
	tm_do_test $h2 interpreted "bla!bla"     [list 0]
	tm_do_test $h2 interpreted "bla!bla!bla" [list 0]
	tm_do_test $h1 interpreted "aap!noot"    [list 1 teun]
	tm_do_test $h1 interpreted "bla!noot"    [list 1 teun]
	
	tm_compile $h2
	tm_compile $h1
	
	tm_do_test $h2 compiled "aap!noot"    [list 1 jet]
	tm_do_test $h2 compiled "bla!noot"    [list 1 jet]
	tm_do_test $h2 compiled "bla!bla"     [list 0]
	tm_do_test $h2 compiled "bla!bla!bla" [list 0]
	tm_do_test $h1 interpreted "aap!noot"    [list 1 teun]
	tm_do_test $h1 interpreted "bla!noot"    [list 1 teun]
}

proc tm_do_test {h cmt s t} {
	set res_t [em_find_string $h $s]
	if {[tm_target_cmp $h $res_t $t] != 0} {
		puts "FAIL $cmt: s ($s) : result ($res_t) : expected ($t)"
	}
}

# =========================================================================
