/*
 * Tracer Yacc grammar
 *
 * ------------------------------------------------------------------------
 */

%{
#include <stdlib.h>
#include <stdarg.h>

#include "parser.h"

/* root node of the parsetree */
spec parsetree;
%}


%token <yt_id>	IDENTIFIER
%token <yt_int>	NUMBER

%token K_DESCRIPTION K_DECLARE K_EXPRESSION K_OBJECTIVE
%token K_ACTION, K_SET K_TO K_INFINITY
%token <yt_casestring>	STRING_LITERAL




%type <yt_spec>		specification
%type <yt_decllist>	declblock decl_list actdecl_list
%type <yt_decl>		decl actdecl
%type <yt_asslist>	assignlist assignblock
%type <yt_assign>	assignment
%type <yt_expr>		expr concat_expr choice_expr replicate_expr
			basic_expr objblock
%type <yt_freq>		freq

%left '.' '>'
%left '|'

%start specification

%%

specification  :	descblock
			declblock
			assignblock
			objblock	{ parsetree  = Spec($2,$3,$4);}
			;

descblock  : K_DESCRIPTION 
	       ;


declblock  :  K_DECLARE decl_list   { $$ = reverse_decllist($2);
										  freelist_decllist($2); }
	       |  K_DECLARE		{ $$ = Nildecllist(); }
	       ;         

decl_list  : decl		{ $$ = Consdecllist($1,Nildecllist()); }
	       | decl ';'	{ $$ = Consdecllist($1,Nildecllist()); }
	       | decl_list decl		{ $$ = Consdecllist($2,$1); }
	       | decl_list decl ';'	{ $$ = Consdecllist($2,$1); }
	       ;

decl	   : K_SET IDENTIFIER
	         '{' actdecl_list '}'
		     { $$ = ASet( $2 , reverse_decllist($4));
		       defoccur( $2, IdTypeDecl($$));
			   freelist_decllist($4);
		     }
	       | actdecl		  { $$ = $1; }
	       ;

actdecl	   : K_ACTION IDENTIFIER STRING_LITERAL
			  { $$ = Action($2, $3);
			    defoccur ($2, IdTypeDecl($$));
			  }
	       ;

actdecl_list   : actdecl	{ $$ = Consdecllist($1,Nildecllist()); }
	       | actdecl ','  { $$ = Consdecllist($1,Nildecllist()); }	
	       | actdecl_list actdecl { $$ = Consdecllist($2,$1); }	
	       | actdecl_list actdecl ','	{ $$ = Consdecllist($2,$1); }
	       | error		{ $$ = Nildecllist();}
	       ;

assignblock: K_EXPRESSION		{ $$ = Nilasslist();}
	       | K_EXPRESSION assignlist { $$ = reverse_asslist($2);
									   freelist_asslist($2); }
	       ;

assignlist : assignment		{ $$ = Consasslist($1,Nilasslist()); }
	       | assignlist assignment	{ $$ = Consasslist($2,$1); }
	       ;

assignment : IDENTIFIER '=' expr ';'
			    {  $$ = Assign($1,$3);
			       defoccur($1 , IdTypeAssign($$));
			    }
	       ;


objblock   : K_OBJECTIVE expr	{$$ = $2; }
	       ;
     
expr	   : replicate_expr '.' concat_expr {
						  $3 = reorder_conc($3);
						  $$ = Nconcat($1,$3);
						  $1->parent=$$;
						  $3->parent=$$;
						}
           | replicate_expr '>' concat_expr {
						  $3 = reorder_conc($3);
						  $$ = Dconcat($1,$3);
						  $1->parent=$$;
						  $3->parent=$$;
						}
	       | replicate_expr '|' choice_expr {
						  $3 = reorder_choice($3);
						  $$ = Choice($1,$3);
						  $1->parent=$$;
						  $3->parent=$$;
						}
	       | replicate_expr { $$ = $1; }
	       ;

concat_expr: concat_expr '.' replicate_expr { $$ = Nconcat($1,$3);
						  $1->parent=$$;
						  $3->parent=$$;
						}
	       | concat_expr '>' replicate_expr { $$ = Dconcat($1,$3); 
						  $1->parent=$$;
						  $3->parent=$$;
						}
	       | replicate_expr			{ $$ = $1; }
	       ;

choice_expr: choice_expr '|' replicate_expr { $$ = Choice($1,$3);
						  $1->parent=$$;
						  $3->parent=$$;
						}
	       | replicate_expr			{ $$ = $1; }
	       ;

replicate_expr : basic_expr			{ $$ = $1; }
	       | basic_expr freq		{ $$ = Repl($1,$2);
						  $1->parent=$$;
						}
	       ;


basic_expr     : '(' expr ')'			{ $$ = $2; }
	       | IDENTIFIER			{ $$ = Const($1);
						  useoccur($1);
						}
	       ; 

freq	       : '[' NUMBER K_TO ']'		{ $$ = LFreq($2); }
	       | '[' NUMBER K_TO K_INFINITY ']'	{ $$ = LFreq($2); }
	       | '[' NUMBER K_TO NUMBER ']'		{ $$ = Freq($2,$4); }
	       | '[' K_TO NUMBER ']'			{ $$ = UFreq($3); }
	       | '[' NUMBER ']'					{ $$ = Freq($2,$2); }
	       | '[' K_INFINITY ']'				{ $$ = IFreq(); }
	       | '*'							{ $$ = LFreq(0); }
	       | '[' K_TO K_INFINITY ']'		{ $$ = LFreq(0); }
	       | '+'							{ $$ = LFreq(1); }
	       | '?'							{ $$ = Freq(0,1); }
       	   ;
           
%%
#include <stdio.h>

extern int column; 
extern int lineno;
extern int nr_errs;

void
nonfatal(int lineno, int column, char *fmt, ... )
{
        va_list args;

        va_start(args, fmt);
        (void)fprintf(stderr, "Error at %s:%d column %d ",filename, lineno, column);
        (void)vfprintf(stderr, fmt, args);
        (void)fprintf(stderr, "\n");
        va_end(args);
        fflush(stderr);
	nr_errs++;
}

void
yyerror(char *s)
{	printf("%s: %s:%d column %d: %s", prog, filename, lineno, column,s);
	printf("\n%s\n", lexerline());
        printf("%*s\n", column, "^");
	nr_errs++;
}

void
warning(char *s)
{	printf("%s: %s:%d column %d: warning: %s", prog, filename, lineno, column,s);
	printf("\n%s\n", lexerline());
        printf("%*s\n", column, "^");
}

