#! @PERL@
#
# $Id: mkprimer.pl,v 1.12.2.13 2006/02/15 11:15:32 belinfan Exp $
#
# AUTHORS
# 	<Axel.Belinfante@cs.utwente.nl>
# 	<jan.feenstra@cs.utwente.nl>
#
# NAME 
#	mkprimer
#
# SYNTAX
#	mkprimer
#		[ -o output_filename [ --language language ] ] 
#		[ --list ] 
#       [ --inputs input_gates ]
#		[ --outputs output_gates ] 
#		[ --config config_file ] 
#		specification_file
#
# DESCRIPTION
#	This perl script generates an executable primer from a known specification
#	for TorX.
#
# OPTIONS
#	--list
#		This option makes mkprimer list the known languages and then exit.
#
#	-o output_filename
#		This option specificies the name of the output_filename, the
#		primer.
#
#	--language language
#		This option specificies the format of the specification_file.
#		It overrules the default file suffix association.
#		This option must be accompanied with the -o option.
#
#	--inputs input_gates
#		This option specifies which input_gates are used.
#		It also enables the generation of the .torx and .gates configuration
#		files.
#
#	--outputs output_gates
#		This option specificies which output_gates are used.
#		It also enables the generation of the .torx and .gates configuration
#		files.
#
#	--config config_file
#		This option specificies the config_file that is to be used.
#
#	specification_file
#		Specifies the name of the specification file.
#
# VERSION
#	1.2
#
# DATE OF LASTCHANGE
#	November 6th, 2000
#
#
# PROGRAMMING DETAILS
# The language specific information are specified in separate modules.
# These modules are located at the directory refered to by $packageScriptDir.
# Each module registers languages and commands by defining two associative 
# arrays, one that maps file suffixes to a language name (convention: 
# language names are all UPPERCASE), and a second one that maps language 
# names to (anonymous!) subroutines that execute the commands necessary to 
# build a primer for that particular language. 
#
# The subroutine called has 9 arguments:
# 1) the complete name of the specification file
# 2) the name of the program that should be the result of the call
# 3) $packageRootDir
# 4) $packageLibDir
# 5) the (path to the) directory of the specification file
# 6) the basename of the specfile (without directory, without suffix) 
# 7) the suffix of the specfile (without directory, without name) 
# 8) the suffix used on the platform for executable files (e.g. .exe)
# 9) the name of an associate array variable containing several parameters,
#    see %params in this file. Current keys in this array are:
#    config, inputs, outputs
#
# Each command is a string that is passed to the ::System routine defined below. 
# This routine will return a string containing the error message if something 
# went wrong, or the empty string otherwise.
#
# Feedback to the user can be given using the ::Msg routine defined below, that 
# takes a strings as argument. (TODO? Make it accept a list of strings?)
#
# The associative arrays mentioned above are registered by invoking ::Register 
# routine that has two references to the associative arrays as arguments:
# 1) the suffix to language mapping
# 2) the language to subroutine mapping
#
# Note: the module should define its own package name (convention for package name: 
# the name of the program/environment that is used to build the primer using the 
# module, in all lowercase), and as last command it should contain 1; to make 
# the Perl require command happy.
#
# EXAMPLE:
#
# -- START Example -----
#	# ======================================
#	package trojka;
#	
#	%suffix = (
#				".trojka"   =>   "PROMELA",
#		     );
#	
#	%lang = (
#		"PROMELA" =>
#		 sub  {
#			my($specfile, $target, $prefix, $libdir,
#			   $specpath, $specname, $specsuffix,
#			   $exesuffix, $paramsref) = @_;
#			my(@cmds) = (
#				"utmake $specfile",
#				"mv utrojka$exesuffix $target$exesuffix",
#			          );
#			
#			::Exec(@cmds);
#		},
#	);
#	
#	::Register(\%lang, \%suffix);
#	
#	1;
#	#
#	# ======================================
# -- END Example -----
#

use strict;

require 5.002;

use Getopt::Long;
use File::Basename;
use File::Find;
use Cwd;

my (
	$prog, $commandline, $result, $usage, $debug, $outputfile, $inputs, $outputs,
    $generategateconfig, $txt, $g, @inputgates, @outputgates, $gatesuffix,
    $packageRootDir, $packageLibDir, $packageBinDir, $packageLibexecDir,
    $packageScriptDir,
	%optcl, $language, $specfile, %knownlang, %knownsuffix, $routine,
	$name, $path, $suffix, $abspath, $relpath, $cwdpath, $prefixtxt, $torxsuffix,
	$outputdir, $outputbasename, $outputabsdir, $outputabsname, $relspecfile,
	$configfile, %params, $list, $EXEEXT,
	$pmdirs, @pmdirs,
   );
   
sub Usage;
sub List;
sub Msg;
sub Register;
sub splitsspecfile;
sub registerprimers;
sub registerprimer;
sub addlang;
sub addsuffix;
sub relative_path;

$| = 1;

# ======================================
#
# Defaults, configuration
#

$prog = 'mkprimer';
$gatesuffix = '.gates';
$torxsuffix = '.torx';
$commandline = join(" ", @ARGV);
$usage = "usage: $prog [-o outputfile [ --language lang ] ] [ --inputs inputgates ] [ --outputs outputgates ] [ --config configfile ] [ --pmdirs dirs ] specfile\n".
  "usage: $prog --list\n".
  "makes an executable program \`\`outputfile'' for specfile\n".
  "outputfile defaults to \"specname\" (basename of \"specfile\" without suffix)\n";

$packageRootDir = "@prefix@";
$packageLibDir = "@libdir@";
$packageBinDir = "@bindir@";
$packageLibexecDir = "@libexecdir@";
$packageScriptDir = "@scriptdir@";

$EXEEXT = "@EXEEXT@";

if ($ENV{TORX_ROOT}) {
	$packageLibDir     =~ s/$packageRootDir/$ENV{TORX_ROOT}/;
	$packageBinDir     =~ s/$packageRootDir/$ENV{TORX_ROOT}/;
	$packageLibexecDir =~ s/$packageRootDir/$ENV{TORX_ROOT}/;
	$packageScriptDir  =~ s/$packageRootDir/$ENV{TORX_ROOT}/;
}

# make sure that we can find scripts/programs in TorX' bindir and libexecdir
$ENV{PATH} = $packageBinDir.":".$packageLibexecDir.":".$ENV{PATH}
	if $ENV{PATH};

# ======================================
#
# command line args parsing
#

%optcl = ();
$result = GetOptions(
			\%optcl,
			"language=s",
			"o=s",
			"config=s",
			"inputs=s",
			"outputs=s",
			"pmdirs=s",
			"list",
			"generategateconfig",
			"debug:i",
		);
die Usage
	unless $result;

$debug = 1 if ((defined $optcl{"debug"}) && ! $optcl{"debug"});
$debug = $optcl{"debug"} if ((defined $optcl{"debug"}) && $optcl{"debug"});
$language = $optcl{"language"} if (defined $optcl{"language"});
$outputfile = $optcl{"o"} if (defined $optcl{"o"});
$configfile = $optcl{"config"} if (defined $optcl{"config"});
$inputs = $optcl{"inputs"} if (defined $optcl{"inputs"});
@inputgates = split(',', $inputs);
$outputs = $optcl{"outputs"} if (defined $optcl{"outputs"});
@outputgates = split(',', $outputs);
$generategateconfig = (defined $optcl{"generategateconfig"})
                     || (defined $optcl{"inputs"})
                     || (defined $optcl{"outputs"})
                     ;
$list = defined $optcl{"list"};

$pmdirs = $optcl{"pmdirs"} if (defined $optcl{"pmdirs"});
@pmdirs = split(',', $pmdirs);

$specfile = $ARGV[0] if $ARGV[0];

die "$prog: Language option requires -o option\n".Usage
	if ($language && ! $outputfile);

# ======================================
#
# Main program
#

# get all language specific associations.
registerprimers;

print STDOUT List and exit if $list;

# Miscellaneous checks
die Usage
	unless $specfile;

($path, $name, $suffix) = splitsspecfile($specfile);
$language = $knownsuffix{$suffix}
	unless $language;
	
$outputfile = $name unless defined $outputfile;

# get full pathname of specfile
$cwdpath = cwd();
chdir($path);
$abspath = cwd();
chdir($cwdpath);
$abspath .= '/' unless $abspath =~ m,/$,;

# get full pathname of outputfile
$outputdir = dirname($outputfile);
$outputbasename = basename($outputfile);
$cwdpath = cwd();
chdir($outputdir);
$outputabsdir = cwd();
chdir($cwdpath);
$outputabsdir .= '/' unless $outputabsdir =~ m,/$,;
$outputabsdir .= '/' unless $outputabsdir =~ m,/$,;
$outputabsname = $outputabsdir . $outputbasename;
$relspecfile = relative_path($outputabsdir, "$abspath$name$suffix");

# set up params associative array
$params{'config'} = $configfile;
$params{'inputs'} = $inputs;
$params{'outputs'} = $outputs;

die "$prog: language not given, and cannot be derived from file suffix\n".Usage
	unless $language;
die "$prog: language unknown: $language\n".Usage
	unless $knownlang{$language};

# We made it through the checks, so we now try to build a primer
Msg "Making $language Primer \"$outputfile\" for $specfile:";
$routine = $knownlang{$language};
&$routine($specfile, $outputfile, $packageRootDir, $packageLibDir,
		  $path, $name, $suffix, $EXEEXT, \%params);
Msg "Making $language Primer \"$outputfile\" for $specfile: Done.";

# generate configuration files etc.
	$prefixtxt = '';
	$prefixtxt .= "# This file is generated by \"$prog\". Do not edit. \n";
	$prefixtxt .= "# It was generated with the command:\n";
	$prefixtxt .= "#     $0 $commandline\n";
	$prefixtxt .= "#\n";
	$prefixtxt .= "\n";
	
	Msg "Generating torx info file \"$outputfile$torxsuffix\":";
	$txt = $prefixtxt;
	$txt .= "SOURCEGIVEN $path$name$suffix\n";
	$txt .= "SOURCEABS $abspath$name$suffix\n";
	$txt .= "SOURCEREL $relspecfile\n";
	$txt .= "SOURCECWD $cwdpath\n";
	Output("$outputfile$torxsuffix", $txt);
	Msg "Generating torx info file \"$outputfile$torxsuffix\": Done.";

if ($generategateconfig) {
	Msg "Generating gate config file \"$outputfile$gatesuffix\":";
	$txt = $prefixtxt;
	foreach $g (@inputgates) {
		$txt .= "INPUT $g\n";
	}
	foreach $g (@outputgates) {
		$txt .= "OUTPUT $g\n";
	}
	Output("$outputfile$gatesuffix", $txt);
	Msg "Generating gate config file \"$outputfile$gatesuffix\": Done.";
}

# ======================================
#
# General Subroutines
#

sub Usage {
	my($l, $s, $result);
	foreach $l (sort keys %knownlang) {
		$result .= "\t$l\t\t\t";
		foreach $s (sort keys %knownsuffix) {
			$result .= " $s"
				if $knownsuffix{$s} eq $l;
		}
		$result .= "\n";
	}
	if ($result) {
		return $usage."recognized languages and suffixes:\n".$result;
	} else {
		return $usage."No languages recognized currently :-(\n";
	}
}

sub List {
	my($l, $s, $result);
	foreach $l (sort keys %knownlang) {
		$result .= "$l\n";
	}
	return $result;
}

sub Msg {
	my($msg) = @_;
	print "$prog: $msg\n";
}

sub Output {
	my($file, $txt) = @_;
	open(O, ">$file")
		or Msg("cannot open gatesfile $file: $!") and return;
	print O $txt;
	close O;
}

sub splitsspecfile {
	my($specfile) = @_;
	my($name, $path, $suffix, @suffixlist);
	@suffixlist = keys(%knownsuffix);
	($name, $path, $suffix) = fileparse($specfile, @suffixlist);
	return ($path, $name, $suffix);
}

sub System {
	my($cmd) = @_;
	my($rc, $result);
	$rc = 0xffff & system($cmd);
	if ($rc == 0) {
		# normal exit
	} elsif ($rc == 0xff00) {
		$result .= "$prog: command failed: $!\n";
	} elsif ($rc > 0x80) {
		$rc >>= 8;
		$result .= "$prog: command returned error status: $rc\n";
	} else {
		$result .= "$prog: command ran with ";
		if ($rc & 0x80) {
			$rc &= ~0x80;
			$result .= "coredump from ";
		}
		$result .= "signal $rc\n";
	}
	return ($result);
}

sub Exec {
	my(@cmds) = @_;
	my($error, $cmd);
	foreach $cmd (@cmds) {
		::Msg("executing \`\`$cmd''");
		$error = ::System($cmd);
		die $error
			if $error;
	}
}

sub relative_path {
	my($fromdir, $tofile) = @_;
	my($commondir,$dotdotpath,$fromcommon,$result);
	# print stderr "mkprimer ($fromdir)($tofile)\n";
	return ""
		if $fromdir eq "";
	$commondir = $fromdir;
	$commondir .= '/'
		unless $commondir =~ m,/$,;
	while ($tofile !~ /^$commondir/) {
		# print stderr "mkprimer: matching ($tofile)($commondir); ($dotdotpath)\n";
		$commondir = dirname($commondir);
		$commondir .= '/'
			unless $commondir =~ m,/$,;
		$dotdotpath .= '/'
			if $dotdotpath ne "";
		$dotdotpath .= '..';
	}
	$fromcommon = $tofile;
	$fromcommon =~ s/^$commondir//;
	$result = $dotdotpath . '/'
		if $dotdotpath ne "";
	$result .= $fromcommon;
	return $result;
}


# ======================================
#
# Load Specification Language Specific Subroutines
#
# The first name.pm will stick and not be overruled by later
# found name.pm because that is how require works.
# This may not be what we want...
# Should we use something else than require?
sub registerprimers {
	my($d);
	find(\&registerprimer, $packageScriptDir);
	foreach $d (reverse(@pmdirs)) {
		find(\&registerprimer, $d);
	}
}

sub doregister {
	my($file) = @_;
	my($result) = eval {require($file)};
	if (! $result) {
		# Msg "Cannot load \"$file\" (contains errors)";
	}
}

sub registerprimer {
	/\.pm$/ && doregister($_);
}

sub addlang {
	my($lang, $routine) = @_;
	$knownlang{$lang} = $routine;
}

sub addsuffix {
	my($suffix, $lang) = @_;
	$knownsuffix{$suffix} = $lang;
}

sub Register {
	my($lref, $sref) = @_;
	my($l, $s);
	foreach $l (keys %$lref) {
		main::addlang($l, $$lref{$l});
	}
	foreach $s (keys %$sref) {
		main::addsuffix($s, $$sref{$s});
	}
}
# ======================================
